# -*- coding: utf-8 -*-
"""Slack class file."""

import re

import slack
from slack.signature import SignatureVerifier

from .bitsdb import Bitsdb
from .client import Client
from .update import Update


class Slack:
    """Slack class."""

    def __init__(
            self,
            token,
            signing_secret=None,
            env="sandbox",
            notifications="#bitsdb-dev"
    ):
        """Initialize an Slack class instance."""
        # connect to slack web client
        self.webclient = slack.WebClient(token=token)

        # signing secret for validating requests
        self.signing_secret = signing_secret

        # set environment and channel for notifications
        self.env = env
        self.notifications = notifications

        # check the token type
        self.token = token
        self.token_type = "unknown"
        if token.startswith("xoxp-"):
            self.token_type = "user"
        elif token.startswith("xoxb-"):
            self.token_type = "bot"

        # create instance of the client class
        self.Bitsdb = Bitsdb
        self.client = Client(self)

    def update(self, args, auth):
        """Return an instance of the update class."""
        return Update(self, args, auth)

    #
    # chat
    #
    def chat_post_message(self, channel, text, **kwargs):
        """Post a message to Slack."""
        return self.webclient.chat_postMessage(
            channel=channel,
            text=text,
            **kwargs
        ).data

    def chat_update(self, channel, timestamp, text, **kwargs):
        """Update a previously-posted Slack message."""
        return self.webclient.chat_update(
            channel=channel,
            text=text,
            ts=timestamp,
            **kwargs,
        )

    #
    # conversations
    #
    def get_channels(self, exclude_archived=True):
        """Return a list of public channels."""
        return self.get_conversations(types="public_channel", exclude_archived=exclude_archived)

    def get_channels_dict(self, exclude_archived=True):
        """Return a dict of public channels."""
        return self.get_conversations_dict(types="public_channel", exclude_archived=exclude_archived)

    def get_conversation(self, channel, include_locale=False, include_num_members=False):
        """Return info about a specific conversation."""
        return self.webclient.conversations_info(
            channel=channel,
            include_locale=include_locale,
            include_num_members=include_num_members,
        ).data

    def get_conversations(self, types="public_channel", exclude_archived=True):
        """Return a list of Slack conversations."""
        conversations = []

        # get first page of results
        response = self.webclient.conversations_list(
            types=types,
            exclude_archived=exclude_archived,
            limit=200,
        )
        conversations.extend(response["channels"])
        cursor = response["response_metadata"]["next_cursor"]

        # get additional pages
        while cursor:
            response = self.webclient.conversations_list(
                types=types,
                exclude_archived=exclude_archived,
                limit=200,
                cursor=cursor,
            )
            conversations.extend(response["channels"])
            cursor = response["response_metadata"]["next_cursor"]

        return conversations

    def get_conversations_dict(self, types="public_channel", exclude_archived=True):
        """Return a dict of Slack conversations by id."""
        conversations = {}
        for conversation in self.get_conversations(types, exclude_archived):
            cid = conversation["id"]
            conversations[cid] = conversation
        return conversations

    def get_groups(self, exclude_archived=True):
        """Return a list of public channels."""
        return self.get_conversations(types="private_channel", exclude_archived=exclude_archived)

    def get_groups_dict(self, exclude_archived=True):
        """Return a doct of public channels."""
        return self.get_conversations_dict(types="private_channel", exclude_archived=exclude_archived)

    def invite_conversation_users(self, channel, users):
        """Invite one or more users to a conversation."""
        return self.webclient.conversations_invite(channel=channel, users=users).data

    def kick_conversation_user(self, channel, user):
        """Kick a user from a conversation."""
        return self.webclient.conversations_kick(channel=channel, user=user).data

    #
    # teams (workspaces)
    #
    def get_team_info(self, team=None):
        """Return the info about the given team (workspace)."""
        return self.webclient.team_info(team=team).data

    def get_team_profile(self, visibility="all"):
        """Return the profile of the current team (workspace)."""
        return self.webclient.team_profile_get(visibility=visibility).data

    #
    # usergroups
    #
    def get_usergroup_users(self, usergroup, include_disabled=False):
        """Return a list of Slack usergroup users."""
        return self.webclient.usergroups_users_list(usergroup=usergroup).get("users", [])

    def get_usergroups(self, include_count=False, include_disabled=False, include_users=False):
        """Return a list of Slack usergroups."""
        return self.webclient.usergroups_list(
            include_count=include_count,
            include_disabled=include_disabled,
            include_users=include_users
        ).get("usergroups", [])

    def get_usergroups_dict(self, include_count=False, include_disabled=False, include_users=False):
        """Return a dict of Slack usergroups by id."""
        usergroups = {}
        for usergroup in self.get_usergroups(
                include_count=include_count,
                include_disabled=include_disabled,
                include_users=include_users
        ):
            uid = usergroup["id"]
            usergroups[uid] = usergroup
        return usergroups

    def update_usergroup_users(self, usergroup, users, include_count=False):
        """Return a list of Slack usergroup users."""
        return self.webclient.usergroups_users_update(
            usergroup=usergroup,
            users=users,
            include_count=False
        ).data

    #
    # users
    #
    def get_user_identity(self):
        """Return info about the logged-in user."""
        return self.webclient.users_identity().data

    def get_user_info(self, user, include_locale=False):
        """Return info about a user."""
        return self.webclient.users_info(include_locale=include_locale).get("user")

    def get_users(self, include_locale=False):
        """Return a list of Slack users."""
        users = []

        # get first page of results
        response = self.webclient.users_list(include_locale=include_locale)
        users.extend(response["members"])
        cursor = response["response_metadata"]["next_cursor"]

        # get additional pages
        while cursor:
            response = self.webclient.users_list(include_locale=include_locale, cursor=cursor)
            users.extend(response["members"])
            cursor = response["response_metadata"]["next_cursor"]

        return users

    def get_users_dict(self, key="id", users=None):
        """Return a dict of Slack users by id."""
        if not users:
            users = self.get_users()

        users_dict = {}
        for user in users:
            k = user[key]
            if k:
                users_dict[k] = user
        return users_dict

    def get_users_by_class(self, users=None):
        """Return a dict of users by class."""
        if not users:
            users = self.get_users()

        # initialize classes
        classes = {
            "bots": [],
            "broad": [],
            "deleted": [],
            "restricted": [],
            "ultra_restricted": [],
            "other": [],
        }

        for user in users:

            # skip deleted users
            if user["deleted"]:
                classes["deleted"].append(user)
                continue

            # get user email address
            email = user["profile"].get("email", "")

            # sort bot users
            if not email:
                classes["bots"].append(user)

            # sort single-channel guests
            elif user["is_ultra_restricted"]:
                classes["ultra_restricted"].append(user)

            # sort multi-channel guests
            elif user["is_restricted"]:
                classes["restricted"].append(user)

            # sort broad users
            elif re.search("@broadinstitute.org", email):
                classes["broad"].append(user)

            # sort others
            else:
                classes["other"].append(user)

        return classes

    def get_users_by_email(self, users=None):
        """Return a dict of slack users."""
        if not users:
            users = self.get_users()

        emails = {}
        for user in users:
            email = user["profile"].get("email")
            if email:
                emails[email] = user
        return emails

    def get_user_profile(self, user=None, include_labels=False):
        """Get the profile for a user."""
        return self.webclient.users_profile_get(
            user=user,
            include_labels=include_labels,
        ).data

    def set_user_profile(self, user, profile=None, name=None, value=None):
        """Get the profile for a user."""
        if profile:
            return self.webclient.users_profile_set(
                user=user,
                profile=profile,
            ).data
        if name and value:
            return self.webclient.users_profile_set(
                user=user,
                name=name,
                value=value,
            ).data
        return None

    #
    # signature verification
    #
    def validate_request(self, request, signing_secret):
        """Validate an incoming Flask request from Slack to make sure it is trusted."""
        body = request.get_data().decode("utf-8")
        headers = request.headers
        return SignatureVerifier(signing_secret).is_valid_request(body, headers)



# -*- coding: utf-8 -*-
"""Slack Update class file."""

import re

from .scim import SCIM


class Update:
    """Slack Update class."""

    def __init__(self, slack, args, auth):
        """Initialize a class instance."""
        self.slack = slack
        self.args = args
        self.auth = auth

        self.email = False
        if "email" in args:
            self.email = True

        self.verbose = False
        if "verbose" in args or auth.verbose:
            self.verbose = True

        self.emplids = None
        self.github_profile_id = None
        self.google_users = None
        self.other_accounts = None
        self.people = None
        self.scim_users = None
        self.slack_emails = None
        self.slack_users = None
        self.team_profile = None

    def _check_user(self, user):
        """Check a single user."""
        user_id = user["id"]
        profile = user["profile"]

        # get profile fields
        email = profile.get("email", "").lower()
        first_name = profile.get("first_name")
        last_name = profile.get("last_name")
        real_name = profile.get("real_name")
        phone = profile.get("phone")
        title = profile.get("title")

        # skip users with no email
        if not email:
            print(f"ERROR: User with no email: {real_name}")
            return None

        # check for an existing other account
        if email in self.other_accounts:
            return None

        # check for an existing person
        if email not in self.people:
            if not user["deleted"]:
                print(f"ERROR: Unknown user: {real_name} <{email}>")
            return None

        # get person
        person = self.people[email]
        person_first_name = person["first_name"]
        person_last_name = person["last_name"]
        person_phone = person.get("primary_work_phone")
        person_title = person["title"]

        suffix = self._get_pronouns_suffix(last_name)
        if suffix:
            person_last_name = f"{person_last_name} {suffix}"

        changes = []

        # check first name
        if person_first_name != first_name:
            changes.append(f"  first_name: {first_name} -> {person_first_name}")
            profile["first_name"] = person_first_name

        # check last name
        if person_last_name != last_name:
            changes.append(f"  last_name: {last_name} -> {person_last_name}")
            profile["last_name"] = person_last_name

        # check phone
        if person_phone and person_phone != phone:
            changes.append(f"  phone: {phone} -> {person_phone}")
            profile["phone"] = person_phone

        # check title
        if person_title != title:
            changes.append(f"  title: {title} -> {person_title}")
            profile["title"] = person_title

        # display changes
        if changes:
            print(f"\n{email}:")
            print("\n".join(changes))
            try:
                self.slack.set_user_profile(user_id, profile)
            except Exception as e:
                print(f"ERROR: Failed to update profile for user: {email} [{e}")

        # check activate
        if not person["terminated"] and user["deleted"]:
            return "activate"

        # check terminated
        if person["terminated"] and not user["deleted"]:
            return "deactivate"

        return None

    def _get_github_profile_field(self):
        """Return the ID of the field that represents the GitHub Profile."""
        label = "GitHub Profile"
        for field in self.team_profile.get("fields", {}):
            if field["label"] == label:
                return field["id"]
        return None

    def _get_google_group_emails(self, email):
        """Return the list of Slack users in a Google group."""
        g = self.auth.google()
        g.auth_service_account(g.scopes, g.subject)
        members = g.directory().get_derived_members(email)
        emails = []
        for member in members:
            if member["type"] != "USER":
                continue
            member_id = member["id"]
            if member_id in self.google_users:
                email = self.google_users[member_id]["primaryEmail"]
            else:
                email = member["email"].lower()
            emails.append(email)
        return emails

    def _get_google_group_users(self, email):
        """Return the list of Slack users in a Google group."""
        g = self.auth.google()
        g.auth_service_account(g.scopes, g.subject)
        members = g.directory().get_derived_members(email)
        users = []
        for member in members:
            if member["type"] != "USER":
                continue
            member_id = member["id"]
            if member_id in self.google_users:
                email = self.google_users[member_id]["primaryEmail"]
            else:
                email = member["email"].lower()
            if email not in self.slack_emails:
                continue
            user = self.slack_emails[email]
            user_id = user["id"]
            # skip deleted and single-channel guest users
            if not user["deleted"] and not user["is_ultra_restricted"] and user_id not in users:
                users.append(user_id)
        return users

    def _get_google_users(self):
        """Return a dict of Google users."""
        g = self.auth.google()
        g.auth_service_account(g.scopes, g.subject)
        fields = "nextPageToken,users(id,primaryEmail,name/fullName,suspended)"
        return g.directory().get_users_dict(fields=fields)

    def _get_manager_user_id(self, person):
        """Return the manager value for scim."""
        manager_user_id = None
        manager_id = person["manager_id"]
        if manager_id in self.emplids:
            manager_email = self.emplids[manager_id]["email_username"]
            if manager_email in self.scim_users:
                manager_user_id = self.scim_users[manager_email]["id"]
        return manager_user_id

    def _get_other_accounts(self):
        """Return a dict of other accounts from BITSdb."""
        mongo = self.auth.mongo()
        other_accounts = {}
        for account in mongo.get_collection("other_accounts"):
            key = account.get("slack_username")
            if not key:
                continue
            key = key.lower()
            if key in other_accounts:
                print(f"ERROR: Duplicate email in other accounts: {key}")
                continue
            other_accounts[key] = account
        return other_accounts

    def _get_slack_user_emails(self):
        """Return a dict of Slack users by email address."""
        if not self.slack_users:
            if self.verbose:
                print("Getting Slack users...")
            self.slack_users = self.slack.get_users_dict()
            if self.verbose:
                print(f"Found {len(self.slack_users)} Slack users.")
        emails = {}
        for user_id in self.slack_users:
            user = self.slack_users[user_id]
            email = user["profile"].get("email")
            if not email:
                continue
            if email in emails:
                print(f"ERROR: Duplicate email: {email}")
                continue
            emails[email.lower()] = user
        self.slack_emails = emails
        return self.slack_emails

    def _get_slack_usergroup_syncs(self):
        """Return a list of Slack Usergroup syncs from BITSdb."""
        emails = []
        usergroup_ids = []
        syncs = []
        for sync in self.auth.mongo().get_collection("slack_usergroup_sync"):
            email = sync["google_group"]
            usergroup_id = sync["slack_usergroup"]
            usergroup_name = sync["slack_usergroup_name"]
            if email in emails:
                print(f"WARNING: Duplicate sync email: {email}")
                continue
            emails.append(email)
            if usergroup_id in usergroup_ids:
                print(f"WARNING: Duplicate sync usergroup: {usergroup_name} [{usergroup_id}]")
                continue
            usergroup_ids.append(usergroup_id)
            syncs.append(sync)
        return syncs

    def _get_people(self):
        """Return a dict of people from BITSdb."""
        collection = "people_people"
        project = "broad-bitsdb-firestore"
        g = self.auth.google()
        firestore = g.firestore(project)
        people = {}
        self.emplids = {}
        for person in firestore.get_collection(collection):
            emplid = person["emplid"]
            self.emplids[emplid] = person
            key = person["email_username"]
            if not key:
                continue
            key = key.lower()
            if key in people:
                print(f"ERROR: Duplicate email in people: {key}")
                continue
            people[key] = person
        return people

    def _get_pronouns_suffix(self, last_name):
        """Return the pronouns suffix for the given last name."""
        suffix = ""
        if last_name and re.search(r"\(.*/.*\)$", last_name):
            suffix = re.sub(r".* (\(.*/.*\))$", r"\1", last_name).replace(" ", "")
        return suffix

    def _prepare_idp_user(self, person):
        """Prepare a user for the Slack IDP."""
        user = {
            "active": not person["terminated"],
            "name": {
                "familyName": person["last_name"],
                "givenName": person["first_name"],
                "honorificPrefix": None,
            },
            "phoneNumbers": [
                {
                    "type": "mobile",
                    "value": None,
                },
                {
                    "primary": True,
                    "value": person["primary_work_phone"],
                }
            ],
            "profileUrl": f"https://people-cloud.broadinstitute.org/people/{person['emplid']}",
            "timezone": "US/New_York",
            "title": person["title"],
            "urn:scim:schemas:extension:enterprise:1.0": {
                "costCenter": None,
                "department": person["department_name"],
                "division": person["org_unit"] or None,
                "employeeNumber": person["emplid"],
                "organization": person["home_institution"],
            },
            "userName": person["username"],
            "userType": person["worker_type"],
        }

        manager_id = self._get_manager_user_id(person)
        if manager_id:
            user["urn:scim:schemas:extension:enterprise:1.0"]["manager"] = {
                "managerId": manager_id
            }

        return user

    def idp_groups(self):
        """Update the IDP Groups in Slack Enterprise Grid."""
        scim = SCIM(self.slack.token)

        if self.verbose:
            print("Getting Slack Usergroups Syncs from BITSdb Mongo DB...")
        syncs = self._get_slack_usergroup_syncs()
        if self.verbose:
            print(f"Usergroup Sync: {len(syncs)}")

        if self.verbose:
            print("Getting Slack Groups from SCIM API...")
        groups = scim.get_groups_dict()
        if self.verbose:
            print(f"Groups: {len(groups)}")

        if self.verbose:
            print("Getting Slack Users from SCIM API...")
        users = scim.get_users_dict(key="email")
        if self.verbose:
            print(f"Users: {len(users)}")

        self.google_users = self._get_google_users()

        sync_groups = []

        # create missing groups and update groups
        for sync in sorted(syncs, key=lambda x: x["google_group"]):

            # get google group email
            google_group_email = sync["google_group"]
            if not google_group_email.endswith("@broadinstitute.org"):
                print(f"Skipping non-Broad group: {google_group_email}")
                continue

            name = google_group_email.replace("@broadinstitute.org", "")

            if name in sync_groups:
                print(f"Duplicate sync group: {name}")
                continue

            # get google group members
            google_group_members = self._get_google_group_emails(google_group_email)

            if self.verbose:
                print(f"Syncing group: {google_group_email}")

            # assemble group member emails into slack users
            google_group_slack_user_ids = []
            for email in google_group_members:
                if email in users:
                    google_group_slack_user_ids.append(users[email]["id"])
            google_group_slack_user_ids = sorted(set(google_group_slack_user_ids))

            # assemble group users into group members
            members = scim.prepare_group_members(google_group_slack_user_ids)

            if name not in groups:
                print(f"Creating group: {name}")
                print(scim.create_group(name=name, members=members))

            else:
                group = groups[name]
                group_member_ids = scim.get_group_member_ids(group)
                if group_member_ids != google_group_slack_user_ids:
                    print(f"Updating group from {len(group_member_ids)} to {len(google_group_slack_user_ids)} members: {name}")
                    group["members"] = members
                    scim.update_group(group["id"], group)

            sync_groups.append(name)

        # find groups to delete
        for name in groups:
            group = groups[name]
            if name not in sync_groups:
                print(f"Deleting group: {name}")
                scim.delete_group(group.id)

    def idp_users(self):
        """Update IDP users in Slack."""
        scim = SCIM(self.slack.token)

        if self.verbose:
            print("Getting Slack Users from SCIM API...")
        self.scim_users = scim.get_users_dict(key="email")
        if self.verbose:
            print(f"Slack Users: {len(self.scim_users)}")

        if self.verbose:
            print("Getting People from Firestore...")
        self.people = self._get_people()
        if self.verbose:
            print(f"People: {len(self.people)}")

        for email in self.scim_users:
            user = self.scim_users[email]

            # fix manager data
            if "urn:scim:schemas:extension:enterprise:1.0" in user:
                if "manager" in user["urn:scim:schemas:extension:enterprise:1.0"]:
                    if not user["urn:scim:schemas:extension:enterprise:1.0"]["manager"].get("managerId"):
                        del user["urn:scim:schemas:extension:enterprise:1.0"]["manager"]

            # skip non-people
            if email not in self.people:
                if email.endswith("@broadinstitute.org"):
                    if self.verbose:
                        print(f"Skipping unknown person: {email}")
                continue

            person = self.people[email]
            idp_user = self._prepare_idp_user(person)

            updates = []
            patch = {}

            for key in sorted(idp_user):
                new = idp_user[key]
                old = user.get(key)
                if new != old:
                    updates.append(f"  {key}: {old} -> {new}")
                    patch[key] = new
                    user[key] = new

            if updates:
                print(f"\nUpdating user: {email}...")
                print("\n".join(updates))
                try:
                    scim.update_user(user["id"], user)
                except Exception as e:
                    print(e)

    def profiles(self):
        """Update user profiles in Slack."""
        if not self.team_profile:
            self.team_profile = self.slack.get_team_profile().get("profile")
        self.github_profile_id = self._get_github_profile_field()

        if not self.people:
            if self.verbose:
                print("Getting people...")
            self.people = self._get_people()
        if self.verbose:
            print(f"Found {len(self.people)} people.\n")

        if not self.slack_users:
            if self.verbose:
                print("Getting Slack users...")
            self.slack_users = self.slack.get_users_dict()
        if self.verbose:
            print(f"Found {len(self.slack_users)} Slack users.")

        for user_id in sorted(self.slack_users, key=lambda x: self.slack_users[x]["profile"].get("email", "")):
            user = self.slack_users[user_id]

            # skip deleted users
            if user["deleted"] or user["is_bot"] or user["id"] == "USLACKBOT":
                continue

            email = user["profile"].get("email")

            # skip people with no email address
            if email not in self.people:
                continue

            # get person and their github login
            person = self.people[email]
            github_login = person.get("github_login")

            # check github profile field
            if github_login and self.github_profile_id:

                # get the user profile with profile fields
                profile = self.slack.get_user_profile(user_id).get("profile")
                fields = profile.get("fields")
                if not fields:
                    fields = {}

                # create the profile value from the github_login
                github_url = f"https://github.com/{github_login}"
                body = {
                    "value": github_url,
                    "alt": github_login,
                }

                if self.github_profile_id not in fields:
                    fields[self.github_profile_id] = body
                elif fields[self.github_profile_id] != body:
                    fields[self.github_profile_id] = body

                if fields != profile.get("fields"):
                    print(f"Updating GitHub Profile for: {email}...")
                    profile["fields"] = fields
                    self.slack.set_user_profile(user_id, profile)

    def usergroups(self):
        """Update usergroups in Slack."""
        self._get_slack_user_emails()

        if self.verbose:
            print("Getting syncs...")
        syncs = self._get_slack_usergroup_syncs()
        if self.verbose:
            print(f"Usergroup Sync Rules: {len(syncs)}")

        if self.verbose:
            print("Getting Usergroups...")
        usergroups = self.slack.get_usergroups_dict(include_disabled=True, include_users=True)
        if self.verbose:
            print(f"Usergroups: {len(usergroups)}")

        self.google_users = self._get_google_users()

        for sync in sorted(syncs, key=lambda x: x.get("google_group", "")):
            google_group = sync.get("google_group")
            usergroup_id = sync.get("slack_usergroup")
            usergroup_name = sync.get("slack_usergroup_name")

            if usergroup_id not in usergroups:
                print(f"ERROR: Usergroup not found: {usergroup_name} [{usergroup_id}] ({google_group})")
                continue

            usergroup = usergroups[usergroup_id]

            # skip deleted usergroups
            if usergroup.get("date_delete"):
                continue

            # get current slack usergroup users
            users = sorted(usergroup.get("users", []))

            # get google group members as slack users
            group_users = sorted(self._get_google_group_users(google_group))

            if users != group_users:
                print(f"\n{google_group} -> {usergroup_name} [{usergroup_id}]")
                add = list(set(group_users) - set(users))
                delete = list(set(users) - set(group_users))
                if add:
                    print(f"  Adding {len(add)} users:")
                    for user_id in sorted(add, key=lambda x: self.slack_users[x]["profile"]["email"].lower()):
                        user = self.slack_users[user_id]
                        email = user["profile"]["email"].lower()
                        print(f"   + {email}")
                if delete:
                    print(f"  Deleting {len(delete)} users:")
                    for user_id in sorted(delete, key=lambda x: self.slack_users[x]["profile"]["email"].lower()):
                        user = self.slack_users[user_id]
                        email = user["profile"]["email"].lower()
                        print(f"   - {email}")
                if add or delete:
                    try:
                        self.slack.update_usergroup_users(usergroup_id, group_users)
                    except Exception as e:
                        print(f"ERROR: Failed to update Usergroup: {usergroup_name} [{usergroup_id}]: {e}")

    def users(self):
        """Update users in Slack."""
        if not self.other_accounts:
            if self.verbose:
                print("Getting other accounts...")
            self.other_accounts = self._get_other_accounts()
        if self.verbose:
            print(f"Found {len(self.other_accounts)} other accounts.\n")

        if not self.people:
            if self.verbose:
                print("Getting people...")
            self.people = self._get_people()
        if self.verbose:
            print(f"Found {len(self.people)} people.\n")

        if not self.slack_users:
            if self.verbose:
                print("Getting Slack users...")
            self.slack_users = self.slack.get_users_dict()
        if self.verbose:
            print(f"Found {len(self.slack_users)} Slack users.")

        bots = []
        restricted = []
        ultra_restricted = []

        activate = []
        deactivate = []

        # check each user
        for user_id in sorted(self.slack_users, key=lambda x: self.slack_users[x]["profile"].get("email", "")):
            user = self.slack_users[user_id]

            # count and skip bot users
            if user["is_bot"] or user["id"] == "USLACKBOT":
                bots.append(user)
                continue

            # count single-channel and multi-channel guests
            if user.get("is_ultra_restricted"):
                ultra_restricted.append(user)
            elif user.get("is_restricted"):
                restricted.append(user)

            # check for users to activate or deactivate
            action = self._check_user(user)
            if action == "activate":
                activate.append(user)
            elif action == "deactivate":
                deactivate.append(user)

        # output activations/decativations
        if activate:
            print(f"\nUsers to activate: {len(activate)}")
            for user in activate:
                print(f"  + {user['profile']['email']}")
        if deactivate:
            print(f"\nUsers to deactivate: {len(deactivate)}")
            for user in deactivate:
                print(f"  - {user['profile']['email']}")

        # output stats about bots/guests
        if self.verbose:
            print(f"\nBot Users: {len(bots)}")
            print(f"Single-Channel Guests: {len(ultra_restricted)}")
            print(f"Multi-Channel Guests: {len(restricted)}")



# -*- coding: utf-8 -*-
"""BITSdb Class for Slack."""


class Bitsdb:
    """GitHub subclass for BITSdb API."""

    def __init__(self):
        """Initialize a class instance."""

    def user(self, u):
        """Return a slack#usergroups#sync for BITSdb API."""
        bitsdb = {
            "kind": "slack#usergroups#sync",
            "id": u.get("id"),
            "user_id": u.get("id"),
            "name": u.get("name"),
            "real_name": u.get("real_name"),
            "team_id": u.get("team_id"),
            "color": u.get("color"),
            "deleted": u.get("deleted"),
            "has_2fa": u.get("has_2fa"),
            "is_app_user": u.get("is_app_user"),
            "is_bot": u.get("is_bot"),
            "is_owner": u.get("is_owner"),
            "is_primary_owner": u.get("is_primary_owner"),
            "is_restricted": u.get("is_restricted"),
            "is_ultra_restricted": u.get("is_ultra_restricted"),
            "tz": u.get("tz"),
            "tz_label": u.get("tz_label"),
            "tz_offset": str(u.get("tz_offset")),
        }
        return bitsdb

    def usergroup(self, u):
        """Return a slack#usergroups#sync for BITSdb API."""
        bitsdb = {
            "kind": "slack#usergroups#sync",
            "id": u["id"],
            "usergroup_id": u["id"],
            "name": u["name"],
            "handle": u["handle"],
            "description": u["description"],
            "user_count": str(u["user_count"]),
        }
        return bitsdb

    def usergroups_sync(self, u):
        """Return a slack#usergroups#sync for BITSdb API."""
        bitsdb = {
            "kind": "slack#usergroups#sync",
            "id": u["slack_usergroup"],
            "google_group_email": u["google_group"],
            "google_group_name": u.get("google_group_name", "NONAME"),
            "slack_usergroup_id": u["slack_usergroup"],
            "slack_usergroup_name": u["slack_usergroup_name"],
        }
        return bitsdb

    def prep(self, collection, data):
        """Return data prepared for BITSdb API."""
        bitsdb = {}
        for oid in sorted(data):
            s = data[oid]
            k = str(oid)
            if collection == "users":
                bitsdb[k] = self.user(s)
            elif collection == "usergroups":
                bitsdb[k] = self.usergroup(s)
            elif collection == "usergroups_syncs":
                bitsdb[k] = self.usergroups_sync(s)
            else:
                print(f"ERROR: Unknown collection \"{collection}\"" % (collection))
        return bitsdb



# -*- coding: utf-8 -*-
"""Slack command line client class."""

import os


class Client:
    """Slack client class."""

    def __init__(self, slack):
        """Initialize a class instance."""
        self.slack = slack
        self.env = slack.env
        self.notifications = slack.notifications
        self.unix_user = os.environ.get("USER", "nobody")

    def channels_list(self):
        """Display a list of channels (public conversations)."""
        slack_channels = self.slack.get_conversations(exclude_archived=False)
        print(f"Public Slack Channels [{len(slack_channels)}]:")
        for channel in sorted(slack_channels, key=lambda x: x["name"]):
            channel_id = channel["id"]
            name = channel["name"]
            if channel["is_archived"]:
                print(f"  * {name} [{channel_id}] (archived)")
            else:
                print(f"  * {name} [{channel_id}]")

    def groups_list(self):
        """Display a list of groups (private conversations)."""
        slack_channels = self.slack.get_conversations(types="private_channel", exclude_archived=False)
        print(f"Private Slack Channels [{len(slack_channels)}]:")
        for channel in sorted(slack_channels, key=lambda x: x["name"]):
            channel_id = channel["id"]
            name = channel["name"]
            if channel["is_archived"]:
                print(f"  * {name} [{channel_id}] (archived)")
            else:
                print(f"  * {name} [{channel_id}]")

    def post_message(self, channel=None, text=None, **kwargs):
        """Post a message to a Slack channel."""
        if not channel:
            channel = self.notifications
        if not text and not kwargs.get("blocks"):
            print("ERROR: no \"text\" and no \"blocks\" included in Slack message post.")
            return None
        if text and self.env and self.unix_user:
            text = f"{text} `[{self.unix_user}@{self.env}]`"
        try:
            return self.slack.chat_post_message(
                channel=channel,
                text=text,
                **kwargs,
            )
        except Exception as error:
            print(f"ERROR: Failed to post message to Slack channel: {channel} [{error}]")
            return None

    def report(self):
        """Display a report of Slack information."""
        channels = self.slack.get_conversations(types="public_channel", exclude_archived=False)
        print(f"Channels: {len(channels)}")

        groups = self.slack.get_conversations(types="private_channel", exclude_archived=False)
        print(f"Groups: {len(groups)}")

        usergroups = self.slack.get_usergroups()
        print(f"Usergroups: {len(usergroups)}")

        users = self.slack.get_users()
        print(f"Users: {len(users)}")

    def usergroups_list(self):
        """Display a list of usergroups."""
        slack_usergroups = self.slack.get_usergroups()
        print(f"Slack Usergroups [{len(slack_usergroups)}]:")
        for usergroup in sorted(slack_usergroups, key=lambda x: x["name"]):
            usergroup_id = usergroup["id"]
            name = usergroup["name"]
            print(f"  * {name} [{usergroup_id}]")

    def users_audit(self):
        """List slack users."""
        print("Getting Slack Users...")
        classes = self.slack.get_users_by_class()

        if classes["other"]:
            print("\nOther [%s]:" % (len(classes["other"])))
            for user in sorted(classes["other"], key=lambda x: x["profile"]["real_name"]):
                print("   %s: %s <%s>" % (
                    user["id"],
                    user["name"],
                    user["profile"]["real_name"],
                ))

        if classes["ultra_restricted"]:
            print("\nSingle-Channel Guests [%s]:" % (len(classes["ultra_restricted"])))
            for user in sorted(classes["ultra_restricted"], key=lambda x: x["profile"]["email"].lower()):
                print(f'  {user["profile"]["email"].lower()}: @{user["name"]} [{user["id"]}]')

        if classes["restricted"]:
            print("\nMulti-Channel Guests [%s]:" % (len(classes["restricted"])))
            for user in sorted(classes["restricted"], key=lambda x: x["profile"]["email"].lower()):
                print(f'  {user["profile"]["email"].lower()}: @{user["name"]} [{user["id"]}]')

        if classes["bots"]:
            print("\nBots [%s]:" % (len(classes["bots"])))
            for user in sorted(classes["bots"], key=lambda x: x["profile"]["real_name"]):
                print(f'  {user["profile"]["real_name"]}: @{user["name"]} [{user["id"]}]')

    def users_list(self):
        """Display a list of users."""
        slack_users = self.slack.get_users()
        print(f"Slack Users [{len(slack_users)}]:")
        for user in sorted(slack_users, key=lambda x: x["name"]):
            user_id = user["id"]
            name = user["name"]
            print(f"  * {name} [{user_id}]")



# -*- coding: utf-8 -*-
"""Slack SCIM client."""

import requests

from slack_scim import Group, SCIMClient


class SCIM:
    """SCIM Class."""

    def __init__(self, token):
        """Initialize a class instance."""
        self.client = SCIMClient(token=token)

        # settings for python requests module
        self.base_url = "https://api.slack.com/scim/v1"
        self.headers = {
            "accept": "application/json",
            "authorization": f"Bearer {token}"
        }

        self.groups = None
        self.groups_dict = None
        self.users = None
        self.users_dict = None

    @classmethod
    def get_group_member_ids(cls, group):
        """Return the primary email address of a SCIM user."""
        member_ids = []
        for member in group["members"]:
            member_ids.append(member["value"])
        return sorted(set(member_ids))

    @classmethod
    def get_user_primary_email(cls, user):
        """Return the primary email address of a SCIM user."""
        for email in user["emails"]:
            if email.get("primary"):
                return email["value"]
        return None

    @classmethod
    def get_user_primary_phone(cls, user):
        """Return the primary phone number of a SCIM user."""
        phone_numbers = user.get("phoneNumbers")
        if phone_numbers is None:
            return None
        for phone in phone_numbers:
            if phone.get("primary"):
                return phone["value"]
        return None

    @classmethod
    def prepare_group_members(cls, user_ids):
        """Return a list of group members, given a list of users."""
        group_members = []
        for user_id in user_ids:
            group_members.append({"value": user_id})
        return group_members

    def create_group(self, name, external_id=None, members=None, users=None):
        """Create an IDP Group."""
        body = {"displayName": name}
        if external_id:
            body["id"] = external_id
        if members:
            body["members"] = members
        elif users:
            body["members"] = self.prepare_group_members(users)
        group: Group = Group.from_dict(body)
        creation_result: Group = self.client.create_group(group)
        return creation_result

    def delete_group(self, group_id):
        """Delete an IDP Group."""
        return self.client.delete_group(group_id)

    def get_group(self, group_id):
        """Return a list of groups."""
        return self.client.read_group(group_id)

    def get_groups(self, search_filter=None):
        """Return a list of groups in Slack Enterprise Grid."""
        url = f"{self.base_url}/Groups"

        count = 1000
        params = {
            "count": count,
            "startIndex": 0,
            "filter": search_filter,
        }

        groups = []
        while True:
            response = requests.get(url, headers=self.headers, params=params).json()
            results = response["Resources"]
            if results:
                groups.extend(results)
            if len(results) < count:
                break
            params["startIndex"] += count
        return groups

    def get_groups_dict(self):
        """Return a dict of SCIM groups by name."""
        groups = {}
        for group in self.get_groups():
            index = group["displayName"]
            groups[index] = group
        return groups

    def get_users_dict(self, key="id"):
        """Return a dict of SCIM users by key."""
        users = {}
        for user in self.get_users():
            if key == "email":
                index = self.get_user_primary_email(user)
            else:
                index = user[key]
            users[index] = user
        return users

    def get_users(self, search_filter=None):
        """Return a list of users in Slack Enterprise Grid."""
        url = f"{self.base_url}/Users"

        count = 1000
        params = {
            "count": count,
            "startIndex": 0,
            "filter": search_filter,
        }

        users = []
        while True:
            response = requests.get(url, headers=self.headers, params=params).json()
            results = response["Resources"]
            if results:
                users.extend(results)
            if len(results) < count:
                break
            params["startIndex"] += count
        return users

    def patch_group(self, group_id, body):
        """Return a group after patching the IDP group."""
        return self.client.patch_group(group_id, body)

    def patch_user(self, user_id, body):
        """Return a user after patching the IDP group."""
        return self.client.patch_user(user_id, body)

    def update_group(self, group_id, group):
        """Overwrite an IDP group."""
        return self.client.update_group(group_id, group)

    def update_user(self, user_id, user):
        """Overwrite an IDP user."""
        return self.client.update_user(user_id, user)
